__all__ = ["Spec"]


# standard library
from dataclasses import dataclass, replace
from dataclasses import Field as Field_, fields as fields_
from functools import lru_cache
from typing import Any, Hashable, List, Optional, Type


# dependencies
from typing_extensions import Literal, get_type_hints


# submodules
from .typing import (
    P,
    AnyDtype,
    AnyName,
    AnyPandas,
    DataClass,
    Role,
    get_dtype,
    get_name,
    get_role,
)


# runtime classes
@dataclass(frozen=True)
class Field:
    """Specification of a field."""

    id: str
    """Identifier of the field."""

    name: AnyName
    """Name of the field."""

    role: Literal["attr", "data", "index"]
    """Role of the field."""

    type: Optional[Any]
    """Type (hint) of the field data."""

    dtype: Optional[AnyDtype]
    """Data type of the field data."""

    default: Any
    """Default value of the field data."""

    @property
    def hashable_name(self) -> Hashable:
        """Hashable name of the field."""
        if isinstance(self.name, dict):
            return tuple(self.name.values())
        else:
            return self.name

    def update(self, obj: DataClass[P]) -> "Field":
        """Update the specification by a dataclass object."""
        return replace(
            self,
            name=format_name(self.name, obj),
            default=getattr(obj, self.id),
        )


class Fields(List[Field]):
    """List of field specifications (with selectors)."""

    @property
    def of_attr(self) -> "Fields":
        """Select only attribute field specifications."""
        return Fields(field for field in self if field.role == "attr")

    @property
    def of_data(self) -> "Fields":
        """Select only data field specifications."""
        return Fields(field for field in self if field.role == "data")

    @property
    def of_index(self) -> "Fields":
        """Select only index field specifications."""
        return Fields(field for field in self if field.role == "index")

    def update(self, obj: DataClass[P]) -> "Fields":
        """Update the specifications by a dataclass object."""
        return Fields(field.update(obj) for field in self)


@dataclass(frozen=True)
class Spec:
    """Specification of a pandas dataclass."""

    fields: Fields
    """List of field specifications."""

    factory: Optional[Type[AnyPandas]] = None
    """Factory for pandas data creation."""

    @classmethod
    def from_dataclass(cls, dataclass: Type[DataClass[P]]) -> "Spec":
        """Create a specification from a data class."""
        fields = Fields()

        for field_ in fields_(eval_types(dataclass)):
            field = convert_field(field_)

            if field is not None:
                fields.append(field)

        factory = getattr(dataclass, "__pandas_factory__", None)
        return cls(fields, factory)

    def update(self, obj: DataClass[P]) -> "Spec":
        """Update the specification by a dataclass object."""
        return replace(self, fields=self.fields.update(obj))

    def __matmul__(self, obj: DataClass[P]) -> "Spec":
        """Alias of the update method."""
        return self.update(obj)


# runtime functions
@lru_cache(maxsize=None)
def convert_field(field_: "Field_[Any]") -> Optional[Field]:
    """Convert a dataclass field to a field specification."""
    role = get_role(field_.type)

    if role is Role.ATTR:
        role = "attr"
    elif role is Role.DATA:
        role = "data"
    elif role is Role.INDEX:
        role = "index"
    else:
        return

    return Field(
        id=field_.name,
        name=get_name(field_.type, field_.name),
        role=role,
        type=field_.type,
        dtype=get_dtype(field_.type),
        default=field_.default,
    )


@lru_cache(maxsize=None)
def eval_types(dataclass: Type[DataClass[P]]) -> Type[DataClass[P]]:
    """Evaluate field types of a dataclass."""
    types = get_type_hints(dataclass, include_extras=True)

    for field in fields_(dataclass):
        field.type = types[field.name]

    return dataclass


def format_name(name: AnyName, obj: DataClass[P]) -> AnyName:
    """Format a name by a dataclass object."""
    if isinstance(name, str):
        return name.format(obj)

    if isinstance(name, dict):
        formatted: "dict[Hashable, Hashable]" = {}

        for key, val in name.items():
            key = key.format(obj) if isinstance(key, str) else key
            val = val.format(obj) if isinstance(val, str) else val
            formatted[key] = val

        return formatted

    return name



__all__ = ["Attr", "Data", "Index", "Name", "Other"]


# standard library
from dataclasses import Field
from enum import Enum, auto
from itertools import chain
from typing import (
    Any,
    ClassVar,
    Collection,
    Hashable,
    Iterable,
    Optional,
    Tuple,
    Type,
    TypeVar,
)


# dependencies
import numpy as np
import pandas as pd
from pandas.api.extensions import ExtensionDtype
from pandas.api.types import pandas_dtype  # type: ignore
from typing_extensions import (
    Annotated,
    Literal,
    ParamSpec,
    Protocol,
    TypeAlias,
    get_args,
    get_origin,
    get_type_hints,
)


# type hints (private)
AnyDtype: TypeAlias = "np.dtype[Any] | ExtensionDtype"
AnyName: TypeAlias = "Hashable | dict[Hashable, Hashable]"
AnyPandas: TypeAlias = "pd.DataFrame | pd.Series"
P = ParamSpec("P")
T = TypeVar("T")
THashable = TypeVar("THashable", bound=Hashable)
TPandas = TypeVar("TPandas", bound=AnyPandas)


class DataClass(Protocol[P]):
    """Type hint for dataclass objects."""

    __dataclass_fields__: ClassVar["dict[str, Field[Any]]"]

    def __init__(self, *args: P.args, **kwargs: P.kwargs) -> None:
        ...


class PandasClass(DataClass[P], Protocol[P, TPandas]):
    """Type hint for dataclass objects with a pandas factory."""

    __pandas_factory__: Type[TPandas]


class Role(Enum):
    """Annotations for typing dataclass fields."""

    ATTR = auto()
    """Annotation for attribute fields."""

    DATA = auto()
    """Annotation for data fields."""

    INDEX = auto()
    """Annotation for index fields."""

    NAME = auto()
    """Annotation for name fields."""

    OTHER = auto()
    """Annotation for other fields."""

    @classmethod
    def annotates(cls, tp: Any) -> bool:
        """Check if any role annotates a type hint."""
        return any(isinstance(arg, cls) for arg in get_args(tp))


# type hints (public)
Attr = Annotated[T, Role.ATTR]
"""Type hint for attribute fields (``Attr[T]``)."""

Data = Annotated[Collection[T], Role.DATA]
"""Type hint for data fields (``Data[T]``)."""

Index = Annotated[Collection[T], Role.INDEX]
"""Type hint for index fields (``Index[T]``)."""

Name = Annotated[THashable, Role.NAME]
"""Type hint for name fields (``Name[T]``)."""

Other = Annotated[T, Role.OTHER]
"""Type hint for other fields (``Other[T]``)."""


# runtime functions
def deannotate(tp: Any) -> Any:
    """Recursively remove annotations in a type hint."""

    class Temporary:
        __annotations__ = dict(tp=tp)

    return get_type_hints(Temporary)["tp"]


def find_annotated(tp: Any) -> Iterable[Any]:
    """Generate all annotated types in a type hint."""
    args = get_args(tp)

    if get_origin(tp) is Annotated:
        yield tp
        yield from find_annotated(args[0])
    else:
        yield from chain(*map(find_annotated, args))


def get_annotated(tp: Any) -> Any:
    """Extract the first role-annotated type."""
    for annotated in filter(Role.annotates, find_annotated(tp)):
        return deannotate(annotated)

    raise TypeError("Could not find any role-annotated type.")


def get_annotations(tp: Any) -> Tuple[Any, ...]:
    """Extract annotations of the first role-annotated type."""
    for annotated in filter(Role.annotates, find_annotated(tp)):
        return get_args(annotated)[1:]

    raise TypeError("Could not find any role-annotated type.")


def get_dtype(tp: Any) -> Optional[AnyDtype]:
    """Extract a NumPy or pandas data type."""
    try:
        dtype = get_args(get_annotated(tp))[0]
    except (IndexError, TypeError):
        return

    if dtype is Any or dtype is type(None):
        return

    if get_origin(dtype) is Literal:
        dtype = get_args(dtype)[0]

    return pandas_dtype(dtype)


def get_name(tp: Any, default: AnyName = None) -> AnyName:
    """Extract a name if found or return given default."""
    try:
        name = get_annotations(tp)[1]
    except (IndexError, TypeError):
        return default

    if isinstance(name, Hashable):
        return name

    if (
        isinstance(name, dict)
        and all(isinstance(key, Hashable) for key in name.keys())
        and all(isinstance(val, Hashable) for val in name.values())
    ):
        return dict(name)

    raise ValueError("Could not find any valid name.")


def get_role(tp: Any, default: Role = Role.OTHER) -> Role:
    """Extract a role if found or return given default."""
    try:
        return get_annotations(tp)[0]
    except TypeError:
        return default



__all__ = ["As", "AsDataFrame", "AsSeries"]


# standard library
from functools import wraps
from types import MethodType
from typing import Any, Callable, Generic, Type, TypeVar


# dependencies
import pandas as pd
from morecopy import copy
from typing_extensions import get_args, get_origin


# submodules
from .parsers import asdataframe, asseries
from .typing import P, AnyPandas, PandasClass


# type hints
TPandas = TypeVar("TPandas", bound=AnyPandas)


# runtime classes
class classproperty:
    """Class property dedicated to ``As.new``."""

    def __init__(self, func: Callable[..., Any]) -> None:
        self.__doc__ = func.__doc__
        self.__func__ = func

    def __get__(
        self,
        obj: Any,
        cls: Type[PandasClass[P, TPandas]],
    ) -> Callable[P, TPandas]:
        return self.__func__(cls)


class As(Generic[TPandas]):
    """Mix-in class that provides shorthand methods."""

    __pandas_factory__: Type[TPandas]
    """Factory for pandas data creation."""

    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Add a pandas factory to an inheriting class."""
        super().__init_subclass__(**kwargs)

        for base in cls.__orig_bases__:  # type: ignore
            if get_origin(base) is As:
                cls.__pandas_factory__ = get_args(base)[0]

    @classproperty
    def new(cls) -> Any:
        """Create a pandas object from dataclass parameters."""
        init = copy(cls.__init__)
        init.__annotations__["return"] = cls.__pandas_factory__

        @wraps(init)
        def new(cls: Any, *args: Any, **kwargs: Any) -> Any:
            if issubclass(cls.__pandas_factory__, pd.DataFrame):
                return asdataframe(cls(*args, **kwargs))
            elif issubclass(cls.__pandas_factory__, pd.Series):
                return asseries(cls(*args, **kwargs))
            else:
                raise TypeError("Not a valid pandas factory.")

        return MethodType(new, cls)


AsDataFrame = As[pd.DataFrame]
"""Alias of ``As[pandas.DataFrame]``."""


AsSeries = As[pd.Series]
"""Alias of ``As[pandas.Series]``."""



__all__ = ["asdataframe", "asseries"]


# standard library
from typing import Any, Hashable, List, Optional, Type, TypeVar, overload


# dependencies
import numpy as np
import pandas as pd


# submodules
from .specs import Spec
from .typing import P, AnyDtype, DataClass, PandasClass


# type hints
TDataFrame = TypeVar("TDataFrame", bound=pd.DataFrame)
TSeries = TypeVar("TSeries", bound=pd.Series)


# runtime functions
@overload
def asdataframe(obj: Any, *, factory: Type[TDataFrame]) -> TDataFrame:
    ...


@overload
def asdataframe(obj: PandasClass[P, TDataFrame], *, factory: None = None) -> TDataFrame:
    ...


@overload
def asdataframe(obj: DataClass[P], *, factory: None = None) -> pd.DataFrame:
    ...


def asdataframe(obj: Any, *, factory: Any = None) -> Any:
    """Create a DataFrame object from a dataclass object."""
    spec = Spec.from_dataclass(type(obj)) @ obj
    attrs = get_attrs(spec)
    data = get_data(spec)
    index = get_index(spec)
    columns = get_columns(spec)

    if factory is None:
        factory = spec.factory or pd.DataFrame

    if not issubclass(factory, pd.DataFrame):
        raise TypeError("Factory must be a subclass of DataFrame.")

    dataframe = factory(data, index, columns)
    dataframe.attrs.update(attrs)
    return dataframe


@overload
def asseries(obj: Any, *, factory: Type[TSeries]) -> TSeries:
    ...


@overload
def asseries(obj: PandasClass[P, TSeries], *, factory: None = None) -> TSeries:
    ...


@overload
def asseries(obj: DataClass[P], *, factory: None = None) -> pd.Series:
    ...


def asseries(obj: Any, *, factory: Any = None) -> Any:
    """Create a Series object from a dataclass object."""
    spec = Spec.from_dataclass(type(obj)) @ obj
    attrs = get_attrs(spec)
    data = get_data(spec)
    index = get_index(spec)

    if data is None:
        name = None
    else:
        name, data = list(data.items())[0]

    if factory is None:
        factory = spec.factory or pd.Series

    if not issubclass(factory, pd.Series):
        raise TypeError("Factory must be a subclass of Series.")

    series = factory(data, index, name=name)
    series.attrs.update(attrs)
    return series


def ensure(data: Any, dtype: Optional[AnyDtype]) -> Any:
    """Ensure data to be 1D and have given data type."""
    if not np.ndim(data):
        data = [data]

    if isinstance(data, (pd.Index, pd.Series)):
        return type(data)(data, dtype=dtype, copy=False)
    else:
        return pd.array(data, dtype=dtype, copy=False)


def get_attrs(spec: Spec) -> "dict[Hashable, Any]":
    """Derive attributes from a specification."""
    attrs: "dict[Hashable, Any]" = {}

    for field in spec.fields.of_attr:
        attrs[field.hashable_name] = field.default

    return attrs


def get_columns(spec: Spec) -> Optional[pd.Index]:
    """Derive columns from a specification."""
    if not spec.fields.of_data:
        return

    names_ = [field.name for field in spec.fields.of_data]

    if all(isinstance(name, Hashable) for name in names_):
        return

    if not all(isinstance(name, dict) for name in names_):
        raise ValueError("All names must be dictionaries.")

    names = [tuple(name.keys()) for name in names_]  # type: ignore
    indexes = [tuple(name.values()) for name in names_]  # type: ignore

    if not len(set(names)) == 1:
        raise ValueError("All name keys must be same.")

    return pd.MultiIndex.from_tuples(indexes, names=names[0])


def get_data(spec: Spec) -> Optional["dict[Hashable, Any]"]:
    """Derive data from a specification."""
    if not spec.fields.of_data:
        return

    names: List[Hashable] = []
    data: List[Any] = []

    for field in spec.fields.of_data:
        names.append(field.hashable_name)
        data.append(ensure(field.default, field.dtype))

    return dict(zip(names, data))


def get_index(spec: Spec) -> Optional[pd.Index]:
    """Derive index from a specification."""
    if not spec.fields.of_index:
        return

    names: List[Hashable] = []
    indexes: List[Any] = []

    for field in spec.fields.of_index:
        names.append(field.hashable_name)
        indexes.append(ensure(field.default, field.dtype))

    indexes = np.broadcast_arrays(*indexes)

    if len(indexes) == 1:
        return pd.Index(indexes[0], name=names[0])
    else:
        return pd.MultiIndex.from_arrays(indexes, names=names)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pandas_dataclasses', 'pandas_dataclasses.core']

package_data = \
{'': ['*']}

install_requires = \
['morecopy>=0.2,<0.3', 'typing-extensions>=4.1,<5.0']

extras_require = \
{':python_full_version >= "3.7.1" and python_version < "3.8"': ['numpy>=1.20,<1.22',
                                                                'pandas>=1.3,<1.4'],
 ':python_version >= "3.8" and python_version < "3.11"': ['numpy>=1.20,<2.0',
                                                          'pandas>=1.3,<2.0']}

setup_kwargs = {
    'name': 'pandas-dataclasses',
    'version': '0.6.0',
    'description': 'pandas data creation made easy by dataclass',
    'long_description': '# pandas-dataclasses\n\n[![Release](https://img.shields.io/pypi/v/pandas-dataclasses?label=Release&color=cornflowerblue&style=flat-square)](https://pypi.org/project/pandas-dataclasses/)\n[![Python](https://img.shields.io/pypi/pyversions/pandas-dataclasses?label=Python&color=cornflowerblue&style=flat-square)](https://pypi.org/project/pandas-dataclasses/)\n[![Downloads](https://img.shields.io/pypi/dm/pandas-dataclasses?label=Downloads&color=cornflowerblue&style=flat-square)](https://pepy.tech/project/pandas-dataclasses)\n[![DOI](https://img.shields.io/badge/DOI-10.5281/zenodo.6127352-cornflowerblue?style=flat-square)](https://doi.org/10.5281/zenodo.6127352)\n[![Tests](https://img.shields.io/github/workflow/status/astropenguin/pandas-dataclasses/Tests?label=Tests&style=flat-square)](https://github.com/astropenguin/pandas-dataclasses/actions)\n\npandas data creation made easy by dataclass\n\n## Overview\n\npandas-dataclass makes it easy to create [pandas] data (Series and DataFrame) by Python\'s [dataclass] that enables to specify their data types, attributes, and names:\n\n<details>\n<summary>Click to see all imports</summary>\n\n```python\nfrom dataclasses import dataclass\nfrom pandas_dataclasses import AsDataFrame, Data, Index\n```\n</details>\n\n```python\n@dataclass\nclass Weather(AsDataFrame):\n    """Weather information."""\n\n    year: Index[int]\n    month: Index[int]\n    temp: Data[float]\n    wind: Data[float]\n\n\ndf = Weather.new(\n    [2020, 2020, 2021, 2021, 2022],\n    [1, 7, 1, 7, 1],\n    [7.1, 24.3, 5.4, 25.9, 4.9],\n    [2.4, 3.1, 2.3, 2.4, 2.6],\n)\n```\n\nwhere `df` will become a DataFrame object like:\n\n```\n            temp  wind\nyear month\n2020 1       7.1   2.4\n     7      24.3   3.1\n2021 1       5.4   2.3\n     7      25.9   2.4\n2022 1       4.9   2.6\n```\n\n### Features\n\n- Type specification of pandas indexes and data\n- Metadata storing in pandas data attributes\n- Support for hierarchical index and columns\n- Support for full [dataclass] features\n- Support for static type check by [Pyright] ([Pylance])\n\n### Installation\n\n```bash\npip install pandas-dataclasses\n```\n\n## How it works\n\npandas-dataclasses provides you the following features:\n\n- Type hints for dataclass fields (`Attr`, `Data`, `Index`) to specify index(es), data, and attributes of pandas data\n- Mix-in classes for dataclasses (`As`, `AsDataFrame`, `AsSeries`) to create pandas data by a classmethod (`new`) that takes the same arguments as dataclass initialization\n\nWhen you call `new`, it will first create a dataclass object and then create a Series or DataFrame object from the dataclass object according the type hints and values in it.\nIn the example above, `df = Weather.new(...)` is thus equivalent to:\n\n```python\nobj = Weather([2020, ...], [1, ...], [7.1, ...], [2.4, ...])\ndf = asdataframe(obj)\n```\n\nwhere `asdataframe` is a conversion function.\npandas-dataclasses does not touch the dataclass object creation itself; this allows you to fully customize your dataclass before conversion by the dataclass features (`field`, `__post_init__`, ...).\n\n## Basic usage\n\n### DataFrame creation\n\nAs shown in the example above, a dataclass that has the `AsDataFrame` mix-in will create DataFrame objects:\n\n<details>\n<summary>Click to see all imports</summary>\n\n```python\nfrom dataclasses import dataclass\nfrom pandas_dataclasses import AsDataFrame, Data, Index\n```\n</details>\n\n```python\n@dataclass\nclass Weather(AsDataFrame):\n    """Weather information."""\n\n    year: Index[int]\n    month: Index[int]\n    temp: Data[float]\n    wind: Data[float]\n\n\ndf = Weather.new(...)\n```\n\nwhere fields typed by `Index` are "index fields", each value of which will become an index or a part of a hierarchical index of a DataFrame object.\nFields typed by `Data` are "data fields", each value of which will become a data column of a DataFrame object.\nFields typed by other types are just ignored in the DataFrame creation.\n\nEach data or index will be cast to the data type specified in a type hint like `Index[int]`.\nUse `Any` or `None` (like `Index[Any]`) if you do not want type casting.\nSee also [data typing rules](#data-typing-rules) for more examples.\n\nBy default, a field name (i.e. an argument name) is used for the name of corresponding data or index.\nSee also [custom naming](#custom-naming) and [naming rules](#naming-rules) if you want customization.\n\n### Series creation\n\nA dataclass that has the `AsSeries` mix-in will create Series objects:\n\n<details>\n<summary>Click to see all imports</summary>\n\n```python\nfrom dataclasses import dataclass\nfrom pandas_dataclasses import AsSeries, Data, Index\n```\n</details>\n\n```python\n@dataclass\nclass Weather(AsSeries):\n    """Weather information."""\n\n    year: Index[int]\n    month: Index[int]\n    temp: Data[float]\n\n\nser = Weather.new(...)\n```\n\nUnlike `AsDataFrame`, the second and subsequent data fields are ignored in the Series creation even if they exist.\nOther rules are the same as for the DataFrame creation.\n\n## Advanced usage\n\n### Metadata storing\n\nFields typed by `Attr` are "attribute fields", each value of which will become an item of attributes of a DataFrame or a Series object:\n\n<details>\n<summary>Click to see all imports</summary>\n\n```python\nfrom dataclasses import dataclass\nfrom pandas_dataclasses import AsDataFrame, Attr, Data, Index\n```\n</details>\n\n```python\n@dataclass\nclass Weather(AsDataFrame):\n    """Weather information."""\n\n    year: Index[int]\n    month: Index[int]\n    temp: Data[float]\n    wind: Data[float]\n    loc: Attr[str] = "Tokyo"\n    lon: Attr[float] = 139.69167\n    lat: Attr[float] = 35.68944\n\n\ndf = Weather.new(...)\n```\n\nwhere `df.attrs` will become like:\n\n```python\n{"loc": "Tokyo", "lon": 139.69167, "lat": 35.68944}\n```\n\n### Custom naming\n\nThe name of data, index, or attribute can be explicitly specified by adding a hashable annotation to the corresponding type:\n\n<details>\n<summary>Click to see all imports</summary>\n\n```python\nfrom dataclasses import dataclass\nfrom typing import Annotated as Ann\nfrom pandas_dataclasses import AsDataFrame, Attr, Data, Index\n```\n</details>\n\n```python\n@dataclass\nclass Weather(AsDataFrame):\n    """Weather information."""\n\n    year: Ann[Index[int], "Year"]\n    month: Ann[Index[int], "Month"]\n    temp: Ann[Data[float], "Temperature (deg C)"]\n    wind: Ann[Data[float], "Wind speed (m/s)"]\n    loc: Ann[Attr[str], "Location"] = "Tokyo"\n    lon: Ann[Attr[float], "Longitude (deg)"] = 139.69167\n    lat: Ann[Attr[float], "Latitude (deg)"] = 35.68944\n\n\ndf = Weather.new(...)\n```\n\nwhere `df` and `df.attrs` will become like:\n\n```\n            Temperature (deg C)  Wind speed (m/s)\nYear Month\n2020 1                      7.1               2.4\n     7                     24.3               3.1\n2021 1                      5.4               2.3\n     7                     25.9               2.4\n2022 1                      4.9               2.6\n```\n\n```python\n{"Location": "Tokyo", "Longitude (deg)": 139.69167, "Latitude (deg)": 35.68944}\n```\n\nAdding dictionary annotations to data fields will create DataFrame objects with hierarchical columns, where dictionary keys will become the names of column levels and dictionary values will become the names of columns:\n\n<details>\n<summary>Click to see all imports</summary>\n\n```python\nfrom dataclasses import dataclass\nfrom typing import Annotated as Ann\nfrom pandas_dataclasses import AsDataFrame, Data, Index\n```\n</details>\n\n```python\ndef name(stat: str, cat: str) -> dict[str, str]:\n    return {"Statistic": stat, "Category": cat}\n\n\n@dataclass\nclass Weather(AsDataFrame):\n    """Weather information."""\n\n    year: Ann[Index[int], "Year"]\n    month: Ann[Index[int], "Month"]\n    temp_avg: Ann[Data[float], name("Temperature (deg C)", "Average")]\n    temp_max: Ann[Data[float], name("Temperature (deg C)", "Maximum")]\n    wind_avg: Ann[Data[float], name("Wind speed (m/s)", "Average")]\n    wind_max: Ann[Data[float], name("Wind speed (m/s)", "Maximum")]\n\n\ndf = Weather.new(...)\n```\n\nwhere `df` will become like:\n\n```\nStatistic  Temperature (deg C)        Wind speed (m/s)\nCategory              Average Maximum          Average Maximum\nYear Month\n2020 1                    7.1    11.1              2.4     8.8\n     7                   24.3    27.7              3.1    10.2\n2021 1                    5.4    10.3              2.3    10.7\n     7                   25.9    30.3              2.4     9.0\n2022 1                    4.9     9.4              2.6     8.8\n```\n\nIf an annotation is a [format string] or a dictionary that has [format string]s as keys and/or values, it will be formatted by a dataclass object before the data creation:\n\n<details>\n<summary>Click to see all imports</summary>\n\n```python\nfrom dataclasses import dataclass\nfrom typing import Annotated as Ann\nfrom pandas_dataclasses import AsDataFrame, Data, Index\n```\n</details>\n\n```python\n@dataclass\nclass Weather(AsDataFrame):\n    """Weather information."""\n\n    year: Ann[Index[int], "Year"]\n    month: Ann[Index[int], "Month"]\n    temp: Ann[Data[float], "Temperature ({.temp_unit})"]\n    wind: Ann[Data[float], "Wind speed ({.wind_unit})"]\n    temp_unit: str = "deg C"\n    wind_unit: str = "m/s"\n```\n\nwhere units of the temperature and the wind speed can be dynamically updated like `Weather.new(..., temp_unit="deg F", wind_unit="km/h"`).\n\n### Custom pandas factory\n\nA custom class can be specified as a factory for the Series or DataFrame creation by `As`, the generic version of `AsDataFrame` and `AsSeries`.\nNote that the custom class must be a subclass of either `pandas.Series` or `pandas.DataFrame`:\n\n<details>\n<summary>Click to see all imports</summary>\n\n```python\nimport pandas as pd\nfrom dataclasses import dataclass\nfrom pandas_dataclasses import As, Data, Index\n```\n</details>\n\n```python\nclass CustomSeries(pd.Series):\n    """Custom pandas Series."""\n\n    pass\n\n\n@dataclass\nclass Temperature(As[CustomSeries]):\n    """Temperature information."""\n\n    year: Index[int]\n    month: Index[int]\n    temp: Data[float]\n\n\nser = Temperature.new(...)\n```\n\nwhere `ser` will be a `CustomSeries` object.\n\n## Appendix\n\n### Data typing rules\n\nThe data type (dtype) of data/index is determined from the first `Data`/`Index` type of the corresponding field.\nThe following table shows how the data type is inferred:\n\n<details>\n<summary>Click to see all imports</summary>\n\n```python\nfrom typing import Any, Annotated as Ann, Literal as L\nfrom pandas_dataclasses import Data\n```\n</details>\n\nType hint | Inferred data type\n--- | ---\n`Data[Any]` | `None` (no type casting)\n`Data[None]` | `None` (no type casting)\n`Data[int]` | `numpy.int64`\n`Data[numpy.int32]` | `numpy.int32`\n`Data[L["datetime64[ns]"]]` | `numpy.dtype("<M8[ns]")`\n`Data[L["category"]]` | `pandas.CategoricalDtype()`\n`Data[int] \\| str` | `numpy.int64`\n`Data[int] \\| Data[float]` | `numpy.int64`\n`Ann[Data[int], "spam"]` | `numpy.int64`\n`Data[Ann[int, "spam"]]` | `numpy.int64`\n\n### Naming rules\n\nThe name of data/index/attribute is determined from the first annotation of the first `Data`/`Index`/`Attr` type of the corresponding field.\nIf the annotation is a [format string] or a dictionary that has [format string]s as keys and/or values, it will be formatted by a dataclass object before the data creation.\nOtherwise, the field name (i.e. argument name) will be used.\nThe following table shows how the name is inferred:\n\n<details>\n<summary>Click to see all imports</summary>\n\n```python\nfrom typing import Any, Annotated as Ann\nfrom pandas_dataclasses import Data\n```\n</details>\n\nType hint | Inferred name\n--- | ---\n`Data[Any]` | (field name)\n`Ann[Data[Any], "spam"]` | `"spam"`\n`Ann[Data[Any], "spam", "ham"]` | `"spam"`\n`Ann[Data[Any], "spam"] \\| Ann[str, "ham"]` | `"spam"`\n`Ann[Data[Any], "spam"] \\| Ann[Data[float], "ham"]` | `"spam"`\n`Ann[Data[Any], "{.name}"` | `"{.name}".format(obj)`\n`Ann[Data[Any], {"0": "spam", "1": "ham"}]` | `("spam", "ham")`\n`Ann[Data[Any], {"0": "{.name}", "1": "ham"}]` | `("{.name}".format(obj), "ham")`\n\nwhere `obj` is a dataclass object that is expected to have `obj.name`.\n\n### Development roadmap\n\nRelease version | Features\n--- | ---\nv0.5 | Support for dynamic naming\nv0.6 | support for extension array and dtype\nv1.0 | Initial major release (freezing public features until v2.0)\n\n<!-- References -->\n[dataclass]: https://docs.python.org/3/library/dataclasses.html\n[format string]: https://docs.python.org/3/library/string.html#format-string-syntax\n[NumPy]: https://numpy.org\n[pandas]: https://pandas.pydata.org\n[Pylance]: https://github.com/microsoft/pylance-release\n[Pyright]: https://github.com/microsoft/pyright\n',
    'author': 'Akio Taniguchi',
    'author_email': 'taniguchi@a.phys.nagoya-u.ac.jp',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/astropenguin/pandas-dataclasses/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7.1,<3.11',
}


setup(**setup_kwargs)
